@@ -1,20 +1,17 @@
 use super::{util, StreamDependency, StreamId};
-use frame::{Error, Frame, Head, Kind};
-use hpack;
+use crate::ext::Protocol;
+use crate::frame::{Error, Frame, Head, Kind};
+use crate::hpack::{self, BytesStr};
 
-use http::{uri, HeaderMap, Method, StatusCode, Uri};
 use http::header::{self, HeaderName, HeaderValue};
+use http::{uri, HeaderMap, Method, Request, StatusCode, Uri};
 
-use byteorder::{BigEndian, ByteOrder};
-use bytes::{Bytes, BytesMut};
-use string::String;
+use bytes::{BufMut, Bytes, BytesMut};
 
 use std::fmt;
 use std::io::Cursor;
 
-// Minimum MAX_FRAME_SIZE is 16kb, so save some arbitrary space for frame
-// head and other header bits.
-const MAX_HEADER_LENGTH: usize = 1024 * 16 - 100;
+type EncodeBuf<'a> = bytes::buf::Limit<&'a mut BytesMut>;
 
 /// Header frame
 ///
@@ -68,9 +65,10 @@ pub struct Continuation {
 pub struct Pseudo {
     // Request
     pub method: Option<Method>,
-    pub scheme: Option<String<Bytes>>,
-    pub authority: Option<String<Bytes>>,
-    pub path: Option<String<Bytes>>,
+    pub scheme: Option<BytesStr>,
+    pub authority: Option<BytesStr>,
+    pub path: Option<BytesStr>,
+    pub protocol: Option<Protocol>,
 
     // Response
     pub status: Option<StatusCode>,
@@ -90,6 +88,9 @@ struct HeaderBlock {
     /// The decoded header fields
     fields: HeaderMap,
 
+    /// Precomputed size of all of our header fields, for perf reasons
+    field_size: usize,
+
     /// Set to true if decoding went over the max header list size.
     is_over_size: bool,
 
@@ -100,11 +101,7 @@ struct HeaderBlock {
 
 #[derive(Debug)]
 struct EncodingHeaderBlock {
-    /// Argument to pass to the HPACK encoder to resume encoding
-    hpack: Option<hpack::EncodeState>,
-
-    /// remaining headers to encode
-    headers: Iter,
+    hpack: Bytes,
 }
 
 const END_STREAM: u8 = 0x1;
@@ -119,12 +116,13 @@ impl Headers {
     /// Create a new HEADERS frame
     pub fn new(stream_id: StreamId, pseudo: Pseudo, fields: HeaderMap) -> Self {
         Headers {
-            stream_id: stream_id,
+            stream_id,
             stream_dep: None,
             header_block: HeaderBlock {
-                fields: fields,
+                field_size: calculate_headermap_size(&fields),
+                fields,
                 is_over_size: false,
-                pseudo: pseudo,
+                pseudo,
             },
             flags: HeadersFlag::default(),
         }
@@ -138,11 +136,12 @@ impl Headers {
             stream_id,
             stream_dep: None,
             header_block: HeaderBlock {
-                fields: fields,
+                field_size: calculate_headermap_size(&fields),
+                fields,
                 is_over_size: false,
                 pseudo: Pseudo::default(),
             },
-            flags: flags,
+            flags,
         }
     }
 
@@ -153,11 +152,15 @@ impl Headers {
         let flags = HeadersFlag(head.flag());
         let mut pad = 0;
 
-        trace!("loading headers; flags={:?}", flags);
+        tracing::trace!("loading headers; flags={:?}", flags);
+
+        if head.stream_id().is_zero() {
+            return Err(Error::InvalidStreamId);
+        }
 
         // Read the padding length
         if flags.is_padded() {
-            if src.len() < 1 {
+            if src.is_empty() {
                 return Err(Error::MalformedMessage);
             }
             pad = src[0] as usize;
@@ -196,19 +199,25 @@ impl Headers {
 
         let headers = Headers {
             stream_id: head.stream_id(),
-            stream_dep: stream_dep,
+            stream_dep,
             header_block: HeaderBlock {
                 fields: HeaderMap::new(),
+                field_size: 0,
                 is_over_size: false,
                 pseudo: Pseudo::default(),
             },
-            flags: flags,
+            flags,
         };
 
         Ok((headers, src))
     }
 
-    pub fn load_hpack(&mut self, src: &mut BytesMut, max_header_list_size: usize, decoder: &mut hpack::Decoder) -> Result<(), Error> {
+    pub fn load_hpack(
+        &mut self,
+        src: &mut BytesMut,
+        max_header_list_size: usize,
+        decoder: &mut hpack::Decoder,
+    ) -> Result<(), Error> {
         self.header_block.load(src, max_header_list_size, decoder)
     }
 
@@ -236,10 +245,6 @@ impl Headers {
         self.header_block.is_over_size
     }
 
-    pub(crate) fn has_too_big_field(&self) -> bool {
-        self.header_block.has_too_big_field()
-    }
-
     pub fn into_parts(self) -> (Pseudo, HeaderMap) {
         (self.header_block.pseudo, self.header_block.fields)
     }
@@ -249,6 +254,11 @@ impl Headers {
         &mut self.header_block.pseudo
     }
 
+    /// Whether it has status 1xx
+    pub(crate) fn is_informational(&self) -> bool {
+        self.header_block.pseudo.is_informational()
+    }
+
     pub fn fields(&self) -> &HeaderMap {
         &self.header_block.fields
     }
@@ -257,16 +267,20 @@ impl Headers {
         self.header_block.fields
     }
 
-    pub fn encode(self, encoder: &mut hpack::Encoder, dst: &mut BytesMut) -> Option<Continuation> {
+    pub fn encode(
+        self,
+        encoder: &mut hpack::Encoder,
+        dst: &mut EncodeBuf<'_>,
+    ) -> Option<Continuation> {
         // At this point, the `is_end_headers` flag should always be set
         debug_assert!(self.flags.is_end_headers());
 
         // Get the HEADERS frame head
         let head = self.head();
 
-        self.header_block.into_encoding()
-            .encode(&head, encoder, dst, |_| {
-            })
+        self.header_block
+            .into_encoding(encoder)
+            .encode(&head, dst, |_| {})
     }
 
     fn head(&self) -> Head {
@@ -287,6 +301,10 @@ impl fmt::Debug for Headers {
             .field("stream_id", &self.stream_id)
             .field("flags", &self.flags);
 
+        if let Some(ref protocol) = self.header_block.pseudo.protocol {
+            builder.field("protocol", protocol);
+        }
+
         if let Some(ref dep) = self.stream_dep {
             builder.field("stream_dep", dep);
         }
@@ -296,9 +314,96 @@ impl fmt::Debug for Headers {
     }
 }
 
+// ===== util =====
+
+#[derive(Debug, PartialEq, Eq)]
+pub struct ParseU64Error;
+
+pub fn parse_u64(src: &[u8]) -> Result<u64, ParseU64Error> {
+    if src.len() > 19 {
+        // At danger for overflow...
+        return Err(ParseU64Error);
+    }
+
+    let mut ret = 0;
+
+    for &d in src {
+        if d < b'0' || d > b'9' {
+            return Err(ParseU64Error);
+        }
+
+        ret *= 10;
+        ret += (d - b'0') as u64;
+    }
+
+    Ok(ret)
+}
+
 // ===== impl PushPromise =====
 
+#[derive(Debug)]
+pub enum PushPromiseHeaderError {
+    InvalidContentLength(Result<u64, ParseU64Error>),
+    NotSafeAndCacheable,
+}
+
 impl PushPromise {
+    pub fn new(
+        stream_id: StreamId,
+        promised_id: StreamId,
+        pseudo: Pseudo,
+        fields: HeaderMap,
+    ) -> Self {
+        PushPromise {
+            flags: PushPromiseFlag::default(),
+            header_block: HeaderBlock {
+                field_size: calculate_headermap_size(&fields),
+                fields,
+                is_over_size: false,
+                pseudo,
+            },
+            promised_id,
+            stream_id,
+        }
+    }
+
+    pub fn validate_request(req: &Request<()>) -> Result<(), PushPromiseHeaderError> {
+        use PushPromiseHeaderError::*;
+        // The spec has some requirements for promised request headers
+        // [https://httpwg.org/specs/rfc7540.html#PushRequests]
+
+        // A promised request "that indicates the presence of a request body
+        // MUST reset the promised stream with a stream error"
+        if let Some(content_length) = req.headers().get(header::CONTENT_LENGTH) {
+            let parsed_length = parse_u64(content_length.as_bytes());
+            if parsed_length != Ok(0) {
+                return Err(InvalidContentLength(parsed_length));
+            }
+        }
+        // "The server MUST include a method in the :method pseudo-header field
+        // that is safe and cacheable"
+        if !Self::safe_and_cacheable(req.method()) {
+            return Err(NotSafeAndCacheable);
+        }
+
+        Ok(())
+    }
+
+    fn safe_and_cacheable(method: &Method) -> bool {
+        // Cacheable: https://httpwg.org/specs/rfc7231.html#cacheable.methods
+        // Safe: https://httpwg.org/specs/rfc7231.html#safe.methods
+        method == Method::GET || method == Method::HEAD
+    }
+
+    pub fn fields(&self) -> &HeaderMap {
+        &self.header_block.fields
+    }
+
+    #[cfg(feature = "unstable")]
+    pub fn into_fields(self) -> HeaderMap {
+        self.header_block.fields
+    }
+
     /// Loads the push promise frame but doesn't actually do HPACK decoding.
     ///
     /// HPACK decoding is done in the `load_hpack` step.
@@ -306,9 +411,13 @@ impl PushPromise {
         let flags = PushPromiseFlag(head.flag());
         let mut pad = 0;
 
+        if head.stream_id().is_zero() {
+            return Err(Error::InvalidStreamId);
+        }
+
         // Read the padding length
         if flags.is_padded() {
-            if src.len() < 1 {
+            if src.is_empty() {
                 return Err(Error::MalformedMessage);
             }
 
@@ -337,19 +446,25 @@ impl PushPromise {
         }
 
         let frame = PushPromise {
-            flags: flags,
+            flags,
             header_block: HeaderBlock {
                 fields: HeaderMap::new(),
+                field_size: 0,
                 is_over_size: false,
                 pseudo: Pseudo::default(),
             },
-            promised_id: promised_id,
+            promised_id,
             stream_id: head.stream_id(),
         };
         Ok((frame, src))
     }
 
-    pub fn load_hpack(&mut self, src: &mut BytesMut, max_header_list_size: usize, decoder: &mut hpack::Decoder) -> Result<(), Error> {
+    pub fn load_hpack(
+        &mut self,
+        src: &mut BytesMut,
+        max_header_list_size: usize,
+        decoder: &mut hpack::Decoder,
+    ) -> Result<(), Error> {
         self.header_block.load(src, max_header_list_size, decoder)
     }
 
@@ -373,62 +488,34 @@ impl PushPromise {
         self.header_block.is_over_size
     }
 
-    pub fn encode(self, encoder: &mut hpack::Encoder, dst: &mut BytesMut) -> Option<Continuation> {
-        use bytes::BufMut;
-
+    pub fn encode(
+        self,
+        encoder: &mut hpack::Encoder,
+        dst: &mut EncodeBuf<'_>,
+    ) -> Option<Continuation> {
         // At this point, the `is_end_headers` flag should always be set
         debug_assert!(self.flags.is_end_headers());
 
         let head = self.head();
         let promised_id = self.promised_id;
 
-        self.header_block.into_encoding()
-            .encode(&head, encoder, dst, |dst| {
-                dst.put_u32_be(promised_id.into());
+        self.header_block
+            .into_encoding(encoder)
+            .encode(&head, dst, |dst| {
+                dst.put_u32(promised_id.into());
             })
     }
 
     fn head(&self) -> Head {
         Head::new(Kind::PushPromise, self.flags.into(), self.stream_id)
     }
-}
 
-impl PushPromise {
     /// Consume `self`, returning the parts of the frame
     pub fn into_parts(self) -> (Pseudo, HeaderMap) {
         (self.header_block.pseudo, self.header_block.fields)
     }
 }
 
-#[cfg(feature = "unstable")]
-impl PushPromise {
-    pub fn new(
-        stream_id: StreamId,
-        promised_id: StreamId,
-        pseudo: Pseudo,
-        fields: HeaderMap,
-    ) -> Self {
-        PushPromise {
-            flags: PushPromiseFlag::default(),
-            header_block: HeaderBlock {
-                fields,
-                is_over_size: false,
-                pseudo,
-            },
-            promised_id,
-            stream_id,
-        }
-    }
-
-    pub fn fields(&self) -> &HeaderMap {
-        &self.header_block.fields
-    }
-
-    pub fn into_fields(self) -> HeaderMap {
-        self.header_block.fields
-    }
-}
-
 impl<T> From<PushPromise> for Frame<T> {
     fn from(src: PushPromise) -> Self {
         Frame::PushPromise(src)
@@ -453,36 +540,39 @@ impl Continuation {
         Head::new(Kind::Continuation, END_HEADERS, self.stream_id)
     }
 
-    pub fn encode(self, encoder: &mut hpack::Encoder, dst: &mut BytesMut) -> Option<Continuation> {
+    pub fn encode(self, dst: &mut EncodeBuf<'_>) -> Option<Continuation> {
         // Get the CONTINUATION frame head
         let head = self.head();
 
-        self.header_block
-            .encode(&head, encoder, dst, |_| {
-            })
+        self.header_block.encode(&head, dst, |_| {})
     }
 }
 
 // ===== impl Pseudo =====
 
 impl Pseudo {
-    pub fn request(method: Method, uri: Uri) -> Self {
+    pub fn request(method: Method, uri: Uri, protocol: Option<Protocol>) -> Self {
         let parts = uri::Parts::from(uri);
 
         let mut path = parts
             .path_and_query
-            .map(|v| v.into())
-            .unwrap_or_else(|| Bytes::new());
+            .map(|v| BytesStr::from(v.as_str()))
+            .unwrap_or(BytesStr::from_static(""));
 
-        if path.is_empty() && method != Method::OPTIONS {
-            path = Bytes::from_static(b"/");
+        match method {
+            Method::OPTIONS | Method::CONNECT => {}
+            _ if path.is_empty() => {
+                path = BytesStr::from_static("/");
+            }
+            _ => {}
         }
 
         let mut pseudo = Pseudo {
             method: Some(method),
             scheme: None,
             authority: None,
-            path: Some(to_string(path)),
+            path: Some(path).filter(|p| !p.is_empty()),
+            protocol,
             status: None,
         };
 
@@ -496,7 +586,7 @@ impl Pseudo {
         // If the URI includes an authority component, add it to the pseudo
         // headers
         if let Some(authority) = parts.authority {
-            pseudo.set_authority(to_string(authority.into()));
+            pseudo.set_authority(BytesStr::from(authority.as_str()));
         }
 
         pseudo
@@ -508,69 +598,87 @@ impl Pseudo {
             scheme: None,
             authority: None,
             path: None,
+            protocol: None,
             status: Some(status),
         }
     }
 
-    pub fn set_scheme(&mut self, scheme: uri::Scheme) {
-        self.scheme = Some(to_string(scheme.into()));
+    #[cfg(feature = "unstable")]
+    pub fn set_status(&mut self, value: StatusCode) {
+        self.status = Some(value);
     }
 
-    pub fn set_authority(&mut self, authority: String<Bytes>) {
+    pub fn set_scheme(&mut self, scheme: uri::Scheme) {
+        let bytes_str = match scheme.as_str() {
+            "http" => BytesStr::from_static("http"),
+            "https" => BytesStr::from_static("https"),
+            s => BytesStr::from(s),
+        };
+        self.scheme = Some(bytes_str);
+    }
+
+    #[cfg(feature = "unstable")]
+    pub fn set_protocol(&mut self, protocol: Protocol) {
+        self.protocol = Some(protocol);
+    }
+
+    pub fn set_authority(&mut self, authority: BytesStr) {
         self.authority = Some(authority);
     }
-}
 
-fn to_string(src: Bytes) -> String<Bytes> {
-    unsafe { String::from_utf8_unchecked(src) }
+    /// Whether it has status 1xx
+    pub(crate) fn is_informational(&self) -> bool {
+        self.status
+            .map_or(false, |status| status.is_informational())
+    }
 }
 
 // ===== impl EncodingHeaderBlock =====
 
 impl EncodingHeaderBlock {
-    fn encode<F>(mut self,
-                 head: &Head,
-                 encoder: &mut hpack::Encoder,
-                 dst: &mut BytesMut,
-                 f: F)
-        -> Option<Continuation>
-    where F: FnOnce(&mut BytesMut),
+    fn encode<F>(mut self, head: &Head, dst: &mut EncodeBuf<'_>, f: F) -> Option<Continuation>
+    where
+        F: FnOnce(&mut EncodeBuf<'_>),
     {
-        let head_pos = dst.len();
+        let head_pos = dst.get_ref().len();
 
         // At this point, we don't know how big the h2 frame will be.
         // So, we write the head with length 0, then write the body, and
         // finally write the length once we know the size.
         head.encode(0, dst);
 
-        let payload_pos = dst.len();
+        let payload_pos = dst.get_ref().len();
 
         f(dst);
 
         // Now, encode the header payload
-        let continuation = match encoder.encode(self.hpack, &mut self.headers, dst) {
-            hpack::Encode::Full => None,
-            hpack::Encode::Partial(state) => Some(Continuation {
+        let continuation = if self.hpack.len() > dst.remaining_mut() {
+            dst.put_slice(&self.hpack.split_to(dst.remaining_mut()));
+
+            Some(Continuation {
                 stream_id: head.stream_id(),
-                header_block: EncodingHeaderBlock {
-                    hpack: Some(state),
-                    headers: self.headers,
-                },
-            }),
+                header_block: self,
+            })
+        } else {
+            dst.put_slice(&self.hpack);
+
+            None
         };
 
         // Compute the header block length
-        let payload_len = (dst.len() - payload_pos) as u64;
+        let payload_len = (dst.get_ref().len() - payload_pos) as u64;
 
         // Write the frame length
-        BigEndian::write_uint(&mut dst[head_pos..head_pos + 3], payload_len, 3);
+        let payload_len_be = payload_len.to_be_bytes();
+        assert!(payload_len_be[0..5].iter().all(|b| *b == 0));
+        (dst.get_mut()[head_pos..head_pos + 3]).copy_from_slice(&payload_len_be[5..]);
 
         if continuation.is_some() {
             // There will be continuation frames, so the `is_end_headers` flag
             // must be unset
-            debug_assert!(dst[head_pos + 4] & END_HEADERS == END_HEADERS);
+            debug_assert!(dst.get_ref()[head_pos + 4] & END_HEADERS == END_HEADERS);
 
-            dst[head_pos + 4] -= END_HEADERS;
+            dst.get_mut()[head_pos + 4] -= END_HEADERS;
         }
 
         continuation
@@ -583,7 +691,7 @@ impl Iterator for Iter {
     type Item = hpack::Header<Option<HeaderName>>;
 
     fn next(&mut self) -> Option<Self::Item> {
-        use hpack::Header::*;
+        use crate::hpack::Header::*;
 
         if let Some(ref mut pseudo) = self.pseudo {
             if let Some(method) = pseudo.method.take() {
@@ -602,6 +710,10 @@ impl Iterator for Iter {
                 return Some(Path(path));
             }
 
+            if let Some(protocol) = pseudo.protocol.take() {
+                return Some(Protocol(protocol));
+            }
+
             if let Some(status) = pseudo.status.take() {
                 return Some(Status(status));
             }
@@ -609,12 +721,9 @@ impl Iterator for Iter {
 
         self.pseudo = None;
 
-        self.fields.next().map(|(name, value)| {
-            Field {
-                name: name,
-                value: value,
-            }
-        })
+        self.fields
+            .next()
+            .map(|(name, value)| Field { name, value })
     }
 }
 
@@ -726,9 +835,13 @@ impl fmt::Debug for PushPromiseFlag {
 
 // ===== HeaderBlock =====
 
-
 impl HeaderBlock {
-    fn load(&mut self, src: &mut BytesMut, max_header_list_size: usize, decoder: &mut hpack::Decoder) -> Result<(), Error> {
+    fn load(
+        &mut self,
+        src: &mut BytesMut,
+        max_header_list_size: usize,
+        decoder: &mut hpack::Decoder,
+    ) -> Result<(), Error> {
         let mut reg = !self.fields.is_empty();
         let mut malformed = false;
         let mut headers_size = self.calculate_header_list_size();
@@ -736,22 +849,23 @@ impl HeaderBlock {
         macro_rules! set_pseudo {
             ($field:ident, $val:expr) => {{
                 if reg {
-                    trace!("load_hpack; header malformed -- pseudo not at head of block");
+                    tracing::trace!("load_hpack; header malformed -- pseudo not at head of block");
                     malformed = true;
                 } else if self.pseudo.$field.is_some() {
-                    trace!("load_hpack; header malformed -- repeated pseudo");
+                    tracing::trace!("load_hpack; header malformed -- repeated pseudo");
                     malformed = true;
                 } else {
                     let __val = $val;
-                    headers_size += decoded_header_size(stringify!($ident).len() + 1, __val.as_str().len());
+                    headers_size +=
+                        decoded_header_size(stringify!($field).len() + 1, __val.as_str().len());
                     if headers_size < max_header_list_size {
                         self.pseudo.$field = Some(__val);
                     } else if !self.is_over_size {
-                        trace!("load_hpack; header list size over max");
+                        tracing::trace!("load_hpack; header list size over max");
                         self.is_over_size = true;
                     }
                 }
-            }}
+            }};
         }
 
         let mut cursor = Cursor::new(src);
@@ -761,13 +875,10 @@ impl HeaderBlock {
         // the hpack state is connection level. In order to maintain correct
         // state for other streams, the hpack decoding process must complete.
         let res = decoder.decode(&mut cursor, |header| {
-            use hpack::Header::*;
+            use crate::hpack::Header::*;
 
             match header {
-                Field {
-                    name,
-                    value,
-                } => {
+                Field { name, value } => {
                     // Connection level header fields are not supported and must
                     // result in a protocol error.
 
@@ -777,51 +888,61 @@ impl HeaderBlock {
                         || name == "keep-alive"
                         || name == "proxy-connection"
                     {
-                        trace!("load_hpack; connection level header");
+                        tracing::trace!("load_hpack; connection level header");
                         malformed = true;
                     } else if name == header::TE && value != "trailers" {
-                        trace!("load_hpack; TE header not set to trailers; val={:?}", value);
+                        tracing::trace!(
+                            "load_hpack; TE header not set to trailers; val={:?}",
+                            value
+                        );
                         malformed = true;
                     } else {
                         reg = true;
 
                         headers_size += decoded_header_size(name.as_str().len(), value.len());
                         if headers_size < max_header_list_size {
+                            self.field_size +=
+                                decoded_header_size(name.as_str().len(), value.len());
                             self.fields.append(name, value);
                         } else if !self.is_over_size {
-                            trace!("load_hpack; header list size over max");
+                            tracing::trace!("load_hpack; header list size over max");
                             self.is_over_size = true;
                         }
                     }
-                },
+                }
                 Authority(v) => set_pseudo!(authority, v),
                 Method(v) => set_pseudo!(method, v),
                 Scheme(v) => set_pseudo!(scheme, v),
                 Path(v) => set_pseudo!(path, v),
+                Protocol(v) => set_pseudo!(protocol, v),
                 Status(v) => set_pseudo!(status, v),
             }
         });
 
         if let Err(e) = res {
-            trace!("hpack decoding error; err={:?}", e);
+            tracing::trace!("hpack decoding error; err={:?}", e);
             return Err(e.into());
         }
 
         if malformed {
-            trace!("malformed message");
-            return Err(Error::MalformedMessage.into());
+            tracing::trace!("malformed message");
+            return Err(Error::MalformedMessage);
         }
 
         Ok(())
     }
 
-    fn into_encoding(self) -> EncodingHeaderBlock {
+    fn into_encoding(self, encoder: &mut hpack::Encoder) -> EncodingHeaderBlock {
+        let mut hpack = BytesMut::new();
+        let headers = Iter {
+            pseudo: Some(self.pseudo),
+            fields: self.fields.into_iter(),
+        };
+
+        encoder.encode(headers, &mut hpack);
+
         EncodingHeaderBlock {
-            hpack: None,
-            headers: Iter {
-                pseudo: Some(self.pseudo),
-                fields: self.fields.into_iter(),
-            },
+            hpack: hpack.freeze(),
         }
     }
 
@@ -834,78 +955,97 @@ impl HeaderBlock {
     /// > overhead of 32 octets for each header field.
     fn calculate_header_list_size(&self) -> usize {
         macro_rules! pseudo_size {
-            ($name:ident) => ({
+            ($name:ident) => {{
                 self.pseudo
                     .$name
                     .as_ref()
                     .map(|m| decoded_header_size(stringify!($name).len() + 1, m.as_str().len()))
                     .unwrap_or(0)
-            });
+            }};
         }
 
-        pseudo_size!(method) +
-        pseudo_size!(scheme) +
-        pseudo_size!(status) +
-        pseudo_size!(authority) +
-        pseudo_size!(path) +
-        self.fields.iter()
-            .map(|(name, value)| decoded_header_size(name.as_str().len(), value.len()))
-            .sum::<usize>()
+        pseudo_size!(method)
+            + pseudo_size!(scheme)
+            + pseudo_size!(status)
+            + pseudo_size!(authority)
+            + pseudo_size!(path)
+            + self.field_size
     }
+}
 
-    /// Iterate over all pseudos and headers to see if any individual pair
-    /// would be too large to encode.
-    pub(crate) fn has_too_big_field(&self) -> bool {
-        macro_rules! pseudo_size {
-            ($name:ident) => ({
-                self.pseudo
-                    .$name
-                    .as_ref()
-                    .map(|m| decoded_header_size(stringify!($name).len() + 1, m.as_str().len()))
-                    .unwrap_or(0)
-            });
-        }
-
-        if pseudo_size!(method) > MAX_HEADER_LENGTH {
-            return true;
-        }
-
-        if pseudo_size!(scheme) > MAX_HEADER_LENGTH {
-            return true;
-        }
-
-        if pseudo_size!(authority) > MAX_HEADER_LENGTH {
-            return true;
-        }
-
-        if pseudo_size!(path) > MAX_HEADER_LENGTH {
-            return true;
-        }
-
-        // skip :status, its never going to be too big
-
-        for (name, value) in &self.fields {
-            if decoded_header_size(name.as_str().len(), value.len()) > MAX_HEADER_LENGTH {
-                return true;
-            }
-        }
-
-        false
-    }
+fn calculate_headermap_size(map: &HeaderMap) -> usize {
+    map.iter()
+        .map(|(name, value)| decoded_header_size(name.as_str().len(), value.len()))
+        .sum::<usize>()
 }
 
 fn decoded_header_size(name: usize, value: usize) -> usize {
     name + value + 32
 }
 
-// Stupid hack to make the set_pseudo! macro happy, since all other values
-// have a method `as_str` except for `String<Bytes>`.
-trait AsStr {
-    fn as_str(&self) -> &str;
-}
+#[cfg(test)]
+mod test {
+    use super::*;
+    use crate::frame;
+    use crate::hpack::{huffman, Encoder};
 
-impl AsStr for String<Bytes> {
-    fn as_str(&self) -> &str {
-        self
+    #[test]
+    fn test_nameless_header_at_resume() {
+        let mut encoder = Encoder::default();
+        let mut dst = BytesMut::new();
+
+        let headers = Headers::new(
+            StreamId::ZERO,
+            Default::default(),
+            HeaderMap::from_iter(vec![
+                (
+                    HeaderName::from_static("hello"),
+                    HeaderValue::from_static("world"),
+                ),
+                (
+                    HeaderName::from_static("hello"),
+                    HeaderValue::from_static("zomg"),
+                ),
+                (
+                    HeaderName::from_static("hello"),
+                    HeaderValue::from_static("sup"),
+                ),
+            ]),
+        );
+
+        let continuation = headers
+            .encode(&mut encoder, &mut (&mut dst).limit(frame::HEADER_LEN + 8))
+            .unwrap();
+
+        assert_eq!(17, dst.len());
+        assert_eq!([0, 0, 8, 1, 0, 0, 0, 0, 0], &dst[0..9]);
+        assert_eq!(&[0x40, 0x80 | 4], &dst[9..11]);
+        assert_eq!("hello", huff_decode(&dst[11..15]));
+        assert_eq!(0x80 | 4, dst[15]);
+
+        let mut world = dst[16..17].to_owned();
+
+        dst.clear();
+
+        assert!(continuation
+            .encode(&mut (&mut dst).limit(frame::HEADER_LEN + 16))
+            .is_none());
+
+        world.extend_from_slice(&dst[9..12]);
+        assert_eq!("world", huff_decode(&world));
+
+        assert_eq!(24, dst.len());
+        assert_eq!([0, 0, 15, 9, 4, 0, 0, 0, 0], &dst[0..9]);
+
+        // // Next is not indexed
+        assert_eq!(&[15, 47, 0x80 | 3], &dst[12..15]);
+        assert_eq!("zomg", huff_decode(&dst[15..18]));
+        assert_eq!(&[15, 47, 0x80 | 3], &dst[18..21]);
+        assert_eq!("sup", huff_decode(&dst[21..]));
+    }
+
+    fn huff_decode(src: &[u8]) -> BytesMut {
+        let mut buf = BytesMut::new();
+        huffman::decode(src, &mut buf).unwrap()
     }
 }
